# -*- coding: utf-8 -*-

require 'date'
require 'active_record'
require 'sidfm_controller_support'

class SidfmRegkeysController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception

  helper :sidfm

  private
  
  include SidfmControllerSupport
  before_action :update_session_updated_at
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license
  before_action :check_roles

  before_action :get_vm_version
  before_action -> { get_summary_table_lines('register_apikey') }

  def check_roles
    @role_admin = @sidfm_roles.include?(:admin)
    unless @role_admin then
      redirect_to home_path, :notice => af_log_user('アクセス権限がありません。') and return
    end
  end

  # ソート
  def sort_apikeys_content(content, order_param)
    col = order_param[0]
    order = order_param[1]

    case col
    when 'name' then
      content = content.order('LOWER(name COLLATE "C") ' + "#{order.upcase}")
    when 'hostnum-limit' then
      content = content.order(%| 
        CASE WHEN hostnum_limit < 0 THEN NULL ELSE hostnum_limit END 
        #{order.upcase}
      |)
      .order('LOWER(name COLLATE "C") ' + "#{order.upcase}")
    when 'hostname-prefix' then
      content = content.order(%|
        CASE WHEN LENGTH(hostname_prefix) = 0 THEN NULL ELSE hostname_prefix END
        #{order.upcase}
      |)
      .order('LOWER(name COLLATE "C") ' + "#{order.upcase}")
    when 'hostname-suffix' then
      content = content.order(%|
        CASE WHEN LENGTH(hostname_suffix) = 0 THEN NULL ELSE hostname_suffix END
        #{order.upcase}
      |)
      .order('LOWER(name COLLATE "C") ' + "#{order.upcase}")
    when 'dup-action' then
      content = content.order(%|
        CASE dup_action
          WHEN 'reject' THEN 1
          WHEN 'use' THEN 2
          WHEN 'num_suffix' THEN 3
          ELSE 4
        END
        #{order.upcase}
      |).order('LOWER(name COLLATE "C") ' + "#{order.upcase}")
    when 'is-dn-orig'
      content = content.order(:is_dn_orig => :"#{order.downcase == 'asc' ? 'desc' : 'asc'}")
                       .order('LOWER(name COLLATE "C") ' + "#{order.upcase}")
    when 'categories' then
      content = content.joins(%|
        LEFT OUTER JOIN (
          SELECT sidfmd_regkey_id, MIN(sidfmd_category_id) AS system_cat_id
          FROM sidfmd_regkey_cats
          GROUP BY sidfmd_regkey_id
        ) AS apikey_system_category
        ON apikey_system_category.sidfmd_regkey_id = sidfmd_regkeys.id
      |).order(%| apikey_system_category.system_cat_id #{order.upcase}|)
        .order('LOWER(name COLLATE "C") ' + "#{order.upcase}")
    when 'period' then
      content = content.order(:begin_date => :"#{order}")
                       .order(:end_date => :"#{order}")
                       .order('LOWER(name COLLATE "C") ' + "#{order.upcase}")
    end

    content = content.order(:id => :"#{order}")
  end

  # フィルタ情報の初期化
  def check_apikeys_content_filters(filters)
    filter_items = [
      { :filter_item => :name, :default => '' },
      { :filter_item => :hostnum_limit, :default => { :num => '', :is_target_unlimit => false } },
      { :filter_item => :hostname_prefix, :default => { :text => '', :is_target_notreg => false } },
      { :filter_item => :hostname_suffix, :default => { :text => '', :is_target_notreg => false } },
      { :filter_item => :dup_action, :default => '' },
      { :filter_item => :is_dn_orig, :default => '' },
      { :filter_item => :categories, :default => { :ids => [], :is_end => false, :is_except => false, :is_notreg => false } },
      { :filter_item => :period, :default => '::' },
    ]

    filter_items.each do |item|
      if !filters[item[:filter_item]] then
        filters[item[:filter_item]] = item[:default]
      end
      if item[:filter_item] == :hostnum_limit then
        if (filters[item[:filter_item]][:num] =~ /\A[0-9]+\z/) == nil then
          filters[item[:filter_item]][:num] = ''
        end
      elsif item[:filter_item] == :categories then
        filters[item[:filter_item]][:ids] = filters[item[:filter_item]][:ids].map(&:to_i)
      end
    end
    filters
  end

  # フィルタ
  def filter_apikeys_content(content, filters)
    #登録キー名検索
    if filters[:name].size > 0 then
      content = content.where("name ILIKE ?", '%' + filters[:name] + '%')
    end

    #ホスト数制限検索
    if filters[:hostnum_limit][:is_target_unlimit] then
      content = content.where(:hostnum_limit => -1)
    else
      if filters[:hostnum_limit][:num].size > 0 then
        content = content.where(:hostnum_limit => filters[:hostnum_limit][:num].to_i)
      end
    end
 
    #接頭辞検索
    if filters[:hostname_prefix][:text].size > 0 or filters[:hostname_prefix][:is_target_notreg] then
      if filters[:hostname_prefix][:text].size > 0 then
        content = content.where("hostname_prefix ILIKE ?", '%' + filters[:hostname_prefix][:text] + '%')
      else
        content = content.where(:hostname_prefix => '')
      end
    end

    #接尾辞検索
    if filters[:hostname_suffix][:text].size > 0 or filters[:hostname_suffix][:is_target_notreg] then
      if filters[:hostname_suffix][:text].size > 0 then
        content = content.where("hostname_suffix ILIKE ?", '%' + filters[:hostname_suffix][:text] + '%')
      else
        content = content.where(:hostname_suffix => '')
      end
    end

    #dup_action検索
    if filters[:dup_action].size > 0 then
      content = content.where(:dup_action => :"#{filters[:dup_action]}")
    end

    #is_dn_orig検索
    if filters[:is_dn_orig].size > 0 then
      content = content.where(:is_dn_orig => (filters[:is_dn_orig] == 'true'))
    end

    #カテゴリ検索
    if filters[:categories][:ids].any? or filters[:categories][:is_notreg] then
      if filters[:categories][:is_except] then
        content = content.where.not(
          :id => SidfmdRegkeyCat.where(:sidfmd_category_id => filters[:categories][:ids])
                                .select(:sidfmd_regkey_id)
        )
      elsif filters[:categories][:is_notreg] then
        content = content.where.not(
          :id => SidfmdRegkeyCat.distinct.select(:sidfmd_regkey_id)
        )
      else
        if filters[:categories][:is_and] then
          content = content.where(
            :id => SidfmdRegkeyCat.group(:sidfmd_regkey_id)
                                  .having(%| ARRAY_AGG(sidfmd_category_id) @> ARRAY[?]|, filters[:categories][:ids])
                                  .select(:sidfmd_regkey_id)
          )
        else
          content = content.where(
            :id => SidfmdRegkeyCat.where(:sidfmd_category_id => filters[:categories][:ids])
                                  .select(:sidfmd_regkey_id)
          )
        end
      end
    end

    #有効期間検索
    if filters[:period] != '::' then
      if /\A(|notset):(|\d{4}\/\d{2}\/\d{2}):(|\d{4}\/\d{2}\/\d{2})\z/ =~ filters[:period] then
        notset = ('notset' == Regexp.last_match(1))
        begindate = Regexp.last_match(2)
        enddate = Regexp.last_match(3)
        if notset then
          #無期限
          content = content.where(:begin_date => nil).where(:end_date => nil)
        else
          begin
            begindate = Date.parse(begindate)
          rescue Exception => e
            begindate = nil
          end
          begin
            enddate = Date.parse(enddate)
          rescue Exception => e
            enddate = nil
          end
          if begindate and enddate then
            content = content.where(%| begin_date >= ? AND end_date <= ?|, begindate,enddate)
          elsif begindate then
            content = content.where(%| begin_date >= ?|, begindate)
          elsif enddate then
            content = content.where(%| end_date <= ?|, enddate)
          else
            #invalid
            filters[:period][:period] = '::'
          end
        end
      end
    end

    content
  end

  public

  def index
    pages_params = params.fetch(:pages, {}).permit(:offset, :per_page, :filters, :order, :checked, :all_checked)
    @apikeys = {}
    @apikeys[:all_content] = SidfmdRegkey.all
    @all_ids = @apikeys[:all_content].pluck(:id)
    @apikeys[:offset] = (pages_params[:offset] || 0).to_i
    @apikeys[:per_page] = (@summary_table_lines || pages_params[:per_page] || 10).to_i
    @apikeys[:order] = (pages_params[:order] || 'name_asc')
    @apikeys[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @apikeys[:all_checked] = ((pages_params[:all_checked] || '') == 'true')
    @apikeys[:filters] = check_apikeys_content_filters( JSON.parse((pages_params[:filters] || '{}'), symbolize_names: true) )
    @apikeys[:content_filtered] = filter_apikeys_content @apikeys[:all_content], @apikeys[:filters]
    @apikeys[:content_filtered] = sort_apikeys_content @apikeys[:content_filtered], @apikeys[:order].split('_')
    @apikeys[:content] = @apikeys[:content_filtered].offset(@apikeys[:offset]).limit(@apikeys[:per_page])
    @apikeys[:total_size] = @apikeys[:content_filtered].count

    @register_host_apikey_item_names = {
      :name => { :name => SidfmdRegkey::NAME_LABEL },
      :api_key => { :name => SidfmdRegkey::APIKEY_LABEL },
      :hostnum_limit => { :name => SidfmdRegkey::HOSTNUM_LIMIT_LABEL },
      :hostname_prefix => { :name => SidfmdRegkey::HOSTNUM_PREFIX_LABEL },
      :hostname_suffix => { :name => SidfmdRegkey::HOSTNUM_SUFFIX_LABEL },
      :dup_action => { :name => SidfmdRegkey::HOSTNUM_DUP_ACTION_LABEL, :modify_name => '同ホスト名が登録済の場合の設定' },
      :is_dn_orig => { :name => SidfmdRegkey::IS_DN_ORIG_LABEL, :modify_name => '表示名の設定' },
      :categories => { :name => SidfmdRegkey::CATEGORIES_LABEL },
      :period => { :name => SidfmdRegkey::PERIOD_LABEL },
    }

    @system_categories = SidfmdCategory.system
    @not_system_categories = SidfmdCategory.not_system #@role_admin ? SidfmdCategory.not_system : SidfmdCategory.where(:id => @user_category_ids)
    @dup_action_names = {
      :reject => SidfmdRegkey::DUP_ACTION_REJECT_LABEL,
      :use => SidfmdRegkey::DUP_ACTION_USE_LABEL,
      :num_suffix => SidfmdRegkey::DUP_ACTION_NUM_SUFFIX_LABEL
    }
  end

  #ホスト登録キー登録
  def create_new_host_apikey
    f_params = params.fetch(:form, {}).permit(:name)
    if !f_params[:name] or f_params[:name].size == 0 then
      redirect_to :action => :index, :notice => '不正な操作です。' and return
    end
    begin
      SidfmdRegkey.register_host_apikey(f_params[:name])
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :index, :notice => 'キーの登録に失敗しました。' and return
    end
    redirect_to :action => :index
  end

  #ホスト登録キー削除
  def delete_host_register_apikey
    f_params = params.fetch(:form, {}).permit(:id)
    if !f_params[:id] then
      redirect_to :action => :index, :notice => '不正な操作です。' and return
    end
    begin
      SidfmdRegkey.delete_host_apikey(f_params[:id])
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :index, :notice => 'キーの削除に失敗しました。' and return
    end
    redirect_to :action => :index
  end

  #ホスト登録キー変更
  def modify_host_apikey
    f_params = params.fetch(:form, {}).permit(
      :target,
      :id,
      :name,
      :update_api_key,
      :remove_api_key,
      :is_unlimit_hostnum,
      :hostnum_limit,
      :hostname_prefix,
      :hostname_suffix,
      :dup_action,
      :is_dn_orig,
      :system_category_id,
      :custom_category_ids,
      :is_unlimit_period,
      :begin_date,
      :end_date,
      :page_info
    )

    begin
      page_info = JSON.parse(f_params[:page_info]).symbolize_keys
    rescue JSON::ParserError => e
      log_exception(e)
      redirect_to :action => :index, :notice => 'APIキーの変更に失敗しました。' and return
    end

    opts = {
      :'pages[offset]' => page_info[:offset],
      :'pages[per_page]' => page_info[:per_page],
      :'pages[order]' => page_info[:order],
      :'pages[filters]' => JSON.generate(page_info[:filters]),
      :'pages[checked]' => JSON.generate(page_info[:checked]),
      :'pages[all_checked]' => page_info[:all_checked]
    }

    if !f_params[:id] or !f_params[:target] then
      redirect_to :action => :index, :notice => '不正な操作です。' and return
    end

    begin
      apikey = SidfmdRegkey.find(f_params[:id].to_i)
    rescue ActiveRecord::RecordNotFound => e
      log_exception(e)
      redirect_to :action => :index, :notice => 'APIキーが削除されています。' and return
    end

    begin
      SidfmdRegkey.transaction do
        case f_params[:target]
        when 'name' then
          unless f_params[:name] then
            redirect_to :action => :index, :notice => '不正な操作です。' and return
          end
          apikey.lock!.update_name f_params[:name]
        when 'api_key' then
          if !f_params[:update_api_key] and !f_params[:remove_api_key] then
            redirect_to :action => :index, :notice => '不正な操作です。' and return
          end
          if f_params[:update_api_key] then
            apikey.lock!.generate_apikey
          else
            apikey.lock!.remove_apikey
          end
        when 'hostnum_limit' then
          if f_params[:is_unlimit_hostnum] then
            apikey.lock!.update_hostnum_limit -1
          else
            num = f_params[:hostnum_limit].to_i
            if num <= 0 then
              redirect_to :action => :index, :notice => '不正な操作です。' and return
            else
              apikey.lock!.update_hostnum_limit num
            end
          end
        when 'hostname_prefix' then
          if !f_params[:hostname_prefix] || (f_params[:hostname_prefix].size > 0 && /\A[0-9a-zA-Z._-]+\z/ !~ f_params[:hostname_prefix]) then
            redirect_to :action => :index, :notice => '不正な操作です。' and return
          end
          apikey.lock!.update_hostname_prefix f_params[:hostname_prefix]
        when 'hostname_suffix' then
          if !f_params[:hostname_suffix] || (f_params[:hostname_suffix].size > 0 && /\A[0-9a-zA-Z._-]+\z/ !~ f_params[:hostname_suffix]) then
            redirect_to :action => :index, :notice => '不正な操作です。' and return
          end
          apikey.lock!.update_hostname_suffix f_params[:hostname_suffix]
        when 'dup_action' then
          unless f_params[:dup_action] then
            redirect_to :action => :index, :notice => '不正な操作です。' and return
          end
          apikey.lock!.update_dup_action f_params[:dup_action]
        when 'is_dn_orig' then
          if !f_params[:is_dn_orig] or (f_params[:is_dn_orig] != 'true' and f_params[:is_dn_orig] != 'false')then
            redirect_to :action => :index, :notice => '不正な操作です。' and return
          end
          apikey.lock!.update_is_dn_orig f_params[:is_dn_orig]
        when 'categories' then
          if !f_params[:system_category_id] or !f_params[:custom_category_ids] then
            redirect_to :action => :index, :notice => '不正な操作です。' and return
          end

          c_category_ids = JSON.parse(f_params[:custom_category_ids])
          c_category_ids = c_category_ids.map(&:to_i)
          s_category_id = f_params[:system_category_id].to_i

          if !SidfmdCategory.where(:is_system => true, :id => s_category_id).exists? or
             !SidfmdCategory.where(:is_system => false, :id => c_category_ids).exists?
          then
            redirect_to :action => :index, :notice => "#{SidfmdRegkey::CATEGORIES_LABEL}の変更に失敗しました。" and return
          end

          category_ids = c_category_ids + [ s_category_id ]
          apikey.lock!.update_categories category_ids
        when 'period' then
          if f_params[:is_unlimit_period] then
            apikey.lock!.update_period nil, nil
          else
            if !f_params[:begin_date] or !f_params[:end_date] then
              redirect_to :action => :index, :notice => '不正な操作です。' and return
            end
            begin_date = f_params[:begin_date].size > 0 ? Date.parse(f_params[:begin_date]) : nil
            end_date = f_params[:end_date].size > 0 ? Date.parse(f_params[:end_date]) : nil
            if begin_date.is_a?(Date) and end_date.is_a?(Date) and begin_date > end_date then
              redirect_to :action => :index, :notice => '不正な操作です。' and return
            end
            apikey.lock!.update_period begin_date, end_date
          end
        else
          logger.error('invalid data')
          redirect_to :action => :index, :notice => '変更に失敗しました。' and return
        end
      end
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :index, :notice => '変更に失敗しました。' and return
    end

    opts[:action] = :index

    redirect_to opts

  end

  #一括削除
  def bulk_delete_host_register_apikey
    f_params = params.fetch(:form, {}).permit(:ids)
    begin 
      ids = JSON.parse(f_params[:ids])
    rescue JSON::ParserError => e
      log_exception(e)
      redirect_to :action => :index, :notice => '不正な操作です。' and return
    end
    begin
      SidfmdRegkey.bulk_delete_host_apikey(ids)
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :index, :notice => '一括削除に失敗しました。' and return
    end
    
    redirect_to :action => :index
  end

  #登録キー名を返す
  def api_get_host_register_apikey_names
    f_params = params.permit(:ids)
    ids = JSON.parse(f_params[:ids])
    names = SidfmdRegkey.where(:id => ids).pluck(:name)
    render :layout => false, :json => names
  end
end
