#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

class SidfmdHostView < ActiveRecord::Base
  belongs_to :sidfmd_host
  has_many :sidfmd_hv_scsids, dependent: :delete_all
  has_many :sidfmd_hv_cvsses, dependent: :delete_all
  has_many :sidfmd_hv_risks, dependent: :delete_all
  has_many :sidfmd_hv_areas, dependent: :delete_all
  has_many :sidfmd_hv_categories, dependent: :delete_all
  has_many :sidfmd_hv_created_ons, dependent: :delete_all
  has_many :sidfmd_hv_due_dates, dependent: :delete_all
  has_many :sidfmd_hv_unf_due_dates, dependent: :delete_all

  scope :user_host_view, -> (is_role_admin, user_host_ids){
    unless is_role_admin then
      where(:sidfmd_host_id => user_host_ids)
    end
  } 

  PROC_UNIT = 50

  # 全データを同期する
  # viewのデータを全て廃棄して、現在のデータを反映する。
  # 時間がかかるので注意
  # 処理中はSidfmdHostViewでテーブルロックがかかることに注意。
  # GUI内の処理では基本的に実行しないこと（メンテナンスページを用意して実行できるようにする可能性はあり）。
  def self.sync_whole_data
    data = SidfmdHost.create_all_content(no_cache: true, category_id: 0, role_admin: true, corresp_required: Sidfm.corresp_required)
    SidfmdHostView.transaction do
      SidfmdHostView.lock
      SidfmdHvUnfDueDate.delete_all
      SidfmdHvDueDate.delete_all
      SidfmdHvCreatedOn.delete_all
      SidfmdHvCategory.delete_all
      SidfmdHvArea.delete_all
      SidfmdHvRisk.delete_all
      SidfmdHvCvss.delete_all
      SidfmdHvScsid.delete_all
      SidfmdHostView.delete_all
      data.each_slice(PROC_UNIT) do |data_s|
        temp_data = []
        data_s.each do |i|
          host = SidfmdHostView.new
          build_row(host, i)
          temp_data << host
        end
        SidfmdHostView.import(temp_data, validate: false, recursive: true) if temp_data.any?
      end
    end
    data
  end

  # 指定した複数のsidfmd_host_idのデータのみ同期する
  # transaction外で実行すること（メソッド内部でtransactionを使っているため）。
  # 指定したidに対応するSidfmdHost, SidfmdHostViewの状態で以下のように動作する。
  #
  # [sidfmd_host_idに対応するSidfmdHost無し]
  # [sidfmd_host_idに対応するSidfmdHostがflg_delete=true]
  # 対応するSidfmdHostViewを削除
  #
  # [sidfmd_host_idに対応するSidfmdHostあり、対応するSidfmdHostView無し]
  # 対応するSidfmdHostViewを新規作成して追加
  #
  # [sidfmd_host_idに対応するSidfmdHost, SidfmdHostView共にあり]
  # 対応するSidfmdHostViewを更新
  def self.sync_data(sidfmd_host_ids = [], ext_host_content = nil)
    # 削除
    not_exist_hostids =
      sidfmd_host_ids - SidfmdHost.where(:id => sidfmd_host_ids).select(:id).distinct.map(&:id)
    delete_hostids = SidfmdHostView.where(:sidfmd_host_id => not_exist_hostids).select(:sidfmd_host_id).distinct.map(&:sidfmd_host_id)
    delete_hostids |=
      SidfmdHostView.
      joins(:sidfmd_host).
      where(:sidfmd_hosts => { :flg_delete => true }).
      where(:sidfmd_host_id => sidfmd_host_ids).
      select(:sidfmd_host_id).
      distinct.
      map(&:sidfmd_host_id)
    delete_hostids.each_slice(PROC_UNIT) do |delete_hostids_s|
      SidfmdHostView.transaction do
        SidfmdHostView.where(:sidfmd_host_id => delete_hostids_s).lock.destroy_all
      end
    end

    modify_hostids = SidfmdHostView.where(:sidfmd_host_id => sidfmd_host_ids).select(:sidfmd_host_id).distinct.map(&:sidfmd_host_id)

    # 追加するhostidを取得
    add_hostids = sidfmd_host_ids - modify_hostids

    # 追加
    no_add_hostids = []
    add_hostids.each_slice(PROC_UNIT) do |add_hostids_s|
      temp_data = []
      host_content = (ext_host_content || SidfmdHost.create_all_content(no_cache: true, category_id: 0, role_admin: true, corresp_required: Sidfm.corresp_required, host_ids: add_hostids_s))
      add_hostids_s.each do |sidfmd_host_id|
        row = host_content.find {|x| x[:sidfmd_host_id] == sidfmd_host_id }
        if row then
          host = SidfmdHostView.new
          build_row(host, row)
          temp_data << host
        else
          no_add_hostids << sidfmd_host_id
        end
      end
      if temp_data.any? then
        SidfmdHostView.transaction do
          SidfmdHostView.import(temp_data, validate: false, recursive: true)
        end
      end
    end
    add_hostids -= no_add_hostids

    # 更新
    no_modify_hostids = []
    modify_hostids.each_slice(PROC_UNIT) do |modify_hostids_s|
      host_content = (ext_host_content || SidfmdHost.create_all_content(no_cache: true, category_id: 0, role_admin: true, corresp_required: Sidfm.corresp_required, host_ids: modify_hostids_s))
      SidfmdHostView.transaction do
        SidfmdHostView.where(:sidfmd_host_id => modify_hostids_s).lock.destroy_all
        temp_data = []
        modify_hostids_s.each do |sidfmd_host_id|
          row = host_content.find {|x| x[:sidfmd_host_id] == sidfmd_host_id }
          if row then
            host = SidfmdHostView.new
            build_row(host, row)
            temp_data << host
          else
            no_modify_hostids << sidfmd_host_id
          end
        end
        if temp_data.any? then
          SidfmdHostView.import(temp_data, validate: false, recursive: true)
          temp_data.clear
        end
        host_content.clear unless ext_host_content
      end
    end
    modify_hostids -= no_modify_hostids

    { :add => add_hostids, :modify => modify_hostids, :delete => delete_hostids }
  end

  # 指定したsidfmd_host_idの配列について、関連するhtml_idの配列を返す。
  # ViewではなくSidfmdIssueから取得しているため、更新前と更新後の2回
  # 取得し、合わせたものをSidfmdVulView.sync_dataに与えてViewを更新する。
  def self.relate_htmlids(hostids = [])
    SidfmdIssue.where(:sidfmd_host_id => hostids).select(:html_id).distinct.map(&:html_id).uniq.compact
  end

  # data.collect {|x| x[:sidfmd_host_id] } に対して、
  # SidfmdVvScsidからduedatesとstatusids及び関連の値を更新する。
  # View以外操作しないこと。
  def self.update_from_rule_action(data)
    corresp_required = Sidfm.corresp_required
    sidfmd_host_ids = data.collect {|x| x[:sidfmd_host_id] }.uniq.compact
    sidfmd_host_ids.each_slice(PROC_UNIT) do |sidfmd_host_ids_s|
      SidfmdHostView.transaction do
        sidfmd_host_ids_s.each do |i|
          row = SidfmdHostView.lock.where(:sidfmd_host_id => i).first
          if row then
            row.sidfmd_hv_scsids.delete_all
            row.sidfmd_hv_due_dates.delete_all
            row.sidfmd_hv_unf_due_dates.delete_all
            max_sri = 0
            min_status_id = SidfmdIssue::STATUS_ALL.max
            max_status_id = SidfmdIssue::STATUS_ALL.min
            max_cvss = -1.0
            dp_f_cnt = 0
            dp_cnt = 0
            fp_f_cnt = 0
            fp_cnt = 0
            min_elapsed_created_on = Time.current
            temp_scsids = []
            temp_due_dates = []
            temp_unf_due_dates = []
            SidfmdVvScsid.where(:sidfmd_host_id => i).each do |j|
              scsid = SidfmdHvScsid.new
              scsid.sidfmd_host_view_id = row.id
              scsid.status_id = j.status_id
              scsid.cvss = (j.cvss || 0.0)
              scsid.sri = j.sri
              temp_scsids << scsid
              due_date = SidfmdHvDueDate.new
              due_date.sidfmd_host_view_id = row.id
              due_date.d = j.due_date
              temp_due_dates << due_date
              min_status_id = j.status_id if min_status_id > j.status_id
              max_status_id = j.status_id if max_status_id < j.status_id
              fp_f_cnt += 1
              if SidfmdIssue::STATUS_UNFINISHED.include?(j.status_id) then
                unf_due_date = SidfmdHvUnfDueDate.new
                unf_due_date.sidfmd_host_view_id = row.id
                unf_due_date.d = j.due_date
                temp_unf_due_dates << unf_due_date
                max_sri = j.sri if max_sri < j.sri
                max_cvss = j.cvss if (j.cvss and max_cvss < j.cvss)
                dp_f_cnt += 1
                dp_cnt += 1 if j.due_date
                min_elapsed_created_on = j.created_on if min_elapsed_created_on < j.created_on
              else
                fp_cnt += 1
              end
            end
            SidfmdHvScsid.import(temp_scsids, validate: false) if temp_scsids.any?
            SidfmdHvDueDate.import(temp_due_dates, validate: false) if temp_due_dates.any?
            SidfmdHvUnfDueDate.import(temp_unf_due_dates, validate: false) if temp_unf_due_dates.any?
            opts = {
              :max_sri => max_sri,
              :min_status_id => min_status_id,
              :max_status_id => max_status_id,
              :min_elapsed_created_on => min_elapsed_created_on,
            }
            opts[:exploitable] = if temp_scsids.any? then
                                   unf_scsids = SidfmdVulView.joins(:sidfmd_vv_scsids)
                                                             .where(:sidfmd_vv_scsids => { :sidfmd_host_id => i, :status_id => SidfmdIssue::STATUS_UNFINISHED })
                                   if unf_scsids.exists? then
                                     unf_scsids.where(:exploitable => true).exists?
                                   else
                                     nil
                                   end
                                 end
            if max_cvss < 0 then
              opts[:max_cvss_score] = nil
              opts[:max_cvss_label] = SidfmdIssueCve.labelling(nil).to_s
            else
              opts[:max_cvss_score] = max_cvss
              opts[:max_cvss_label] = SidfmdIssueCve.labelling(max_cvss).to_s
            end
            opts[:warning_orig_sri] = (max_sri >= corresp_required[:sri])
            opts[:warning_orig_cvss] = (max_cvss >= corresp_required[:cvss])
            opts[:warning_sri] = SidfmdIssue::STATUS_UNFINISHED.include?(min_status_id) ? opts[:warning_orig_sri] : false
            opts[:warning_cvss] = SidfmdIssue::STATUS_UNFINISHED.include?(min_status_id) ? opts[:warning_orig_cvss] : false
            opts[:fixed_warning_sri] = (SidfmdIssue::STATUS_FINISHED.include?(min_status_id) & opts[:warning_orig_sri])
            opts[:fixed_warning_cvss] = (SidfmdIssue::STATUS_FINISHED.include?(min_status_id) & opts[:warning_orig_cvss])
            opts[:not_fixed_warning_sri] = opts[:warning_sri]
            opts[:not_fixed_warning_cvss] = opts[:warning_cvss]
            if dp_f_cnt > 0 then
              opts[:duedates_percent] = dp_cnt.to_f / dp_f_cnt.to_f * 100.0
            else
              opts[:duedates_percent] = 0.0
            end
            if fp_f_cnt > 0 then
              opts[:fixed_percent] = fp_cnt.to_f / fp_f_cnt.to_f * 100.0
            else
              opts[:fixed_percent] = 0.0
            end
            if min_elapsed_created_on then
              elapsed_days = ((Time.current - min_elapsed_created_on) / 24.0 / 60.0 / 60.0).to_i
              opts[:elapsed_days] = elapsed_days
            else
              opts[:elapsed_days] = -1
            end
            row.update_attributes opts if opts.size > 0
          end
        end
      end
    end
  end

  private

  public

  def self.build_row(host, row, modify = false)
    host.kb = row[:kb]
    host.sidfmd_host_id = row[:sidfmd_host_id]
    host.hostname = row[:hostname]
    host.ext_name = row[:ext_name]
    host.display_name = row[:display_name]
    host.last_checked_at = row[:last_checked_at]
    host.exploitable = row[:exploitable]
    host.max_sri = row[:max_sri]
    host.max_area_id = row[:max_area_id]
    host.min_status_id = row[:min_status_id]
    host.max_status_id = row[:max_status_id]
    host.elapsed_days = row[:elapsed_days]
    host.min_elapsed_created_on = row[:elapsed_created_ons].min
    if row[:os] and row[:os].size > 0 then
      host.osname = row[:os][:osname]
      if row[:os][:dist] then
        host.queue_dist_id = row[:os][:dist].id
      elsif row[:os][:app] and row[:os][:app].size > 0 then
        app = row[:os][:app]
        app = row[:os][:app].first if app.kind_of?(Array)
        if app[:product] then
          host.product_id = app[:product].pid
        end
        if app[:version] then
          host.version_id = app[:version].vid
          unless host.product_id then
            begin
              host.product_id = app[:version].sidfmd_product.pid
            rescue Exception
            end
          end
        end
      end
    else
      host.product_id = nil
      host.version_id = nil
      host.queue_dist_id = nil
      host.osname = nil
    end
    if row[:system_category] and row[:system_category].kind_of?(Hash) then
      host.system_category_id = row[:system_category][:id]
      host.system_category_name = row[:system_category][:name]
    end
    if row[:max_cvss] and row[:max_cvss].kind_of?(Hash) then
      host.max_cvss_exploitable = row[:max_cvss][:exploitable]
      host.max_cvss_score = row[:max_cvss][:score]
      host.max_cvss_label = row[:max_cvss][:label].to_s
    end
    [ 'warning_orig', 'warning', 'fixed_warning', 'not_fixed_warning' ].each do |warning|
      w_sym = warning.to_sym
      if row[w_sym] and row[w_sym].kind_of?(Hash) then
        host.send((warning + '_sri=').to_sym, row[w_sym][:sri])
        host.send((warning + '_cvss=').to_sym, row[w_sym][:cvss])
      end
    end
    if row[:due_dates_unfinished] then
      f_cnt = row[:due_dates_unfinished].size
      if f_cnt > 0 and row[:due_dates_unfinished].compact.size > 0 then
        cnt = row[:due_dates_unfinished].inject(0) {|r, x| r += 1 if (x and x.size > 0); r }
        host.duedates_percent = cnt.to_f / f_cnt.to_f * 100.0
      else
        host.duedates_percent = 0.0
      end
    end
    if row[:status_ids] then
      f_cnt = row[:status_ids].size
      if f_cnt > 0 and row[:status_ids].compact.size > 0 then
        cnt = row[:status_ids].inject(0) {|r, x|
          r += 1 if SidfmdIssue::STATUS_FINISHED.include?(x)
          r
        }
        host.fixed_percent = cnt.to_f / f_cnt.to_f * 100.0
      else
        host.fixed_percent = 0.0
      end
    end
    if modify then
      host.save
    end

    if modify then
      host.sidfmd_hv_scsids.delete_all
      hv_scsids = []
      row[:sri_cvss_status_ids].each do |j|
        scsid = SidfmdHvScsid.new
        scsid.sidfmd_host_view_id = host.id
        scsid.status_id = j[:status_id]
        scsid.cvss = j[:cvss]
        scsid.sri = j[:sri]
        hv_scsids << scsid
      end
      SidfmdHvScsid.import(hv_scsids, validate: false) if hv_scsids.any?
    else
      row[:sri_cvss_status_ids].each do |j|
        host.sidfmd_hv_scsids.build(status_id: j[:status_id],
                                    cvss: j[:cvss],
                                    sri: j[:sri])
      end
    end

    if modify then
      host.sidfmd_hv_cvsses.delete_all
      hv_cvsses = []
      row[:cvss].each do |j|
        cvss = SidfmdHvCvss.new
        cvss.sidfmd_host_view_id = host.id
        cvss.exploitable = j[:exploitable]
        cvss.score = j[:score]
        cvss.label = j[:label].to_s
        hv_cvsses << cvss
      end
      SidfmdHvCvss.import(hv_cvsses, validate: false) if hv_cvsses.any?
    else
      row[:cvss].each do |j|
        host.sidfmd_hv_cvsses.build(exploitable: j[:exploitable],
                                    score: j[:score],
                                    label: j[:label].to_s)
      end
    end

    if modify then
      host.sidfmd_hv_categories.delete_all
      hv_categories = []
      row[:no_system_categories].each do |j|
        category = SidfmdHvCategory.new
        category.sidfmd_host_view_id = host.id
        category.sidfmd_category_id = j[:id]
        category.name = j[:name]
        hv_categories << category
      end
      SidfmdHvCategory.import(hv_categories, validate: false) if hv_categories.any?
    else
      row[:no_system_categories].each do |j|
        host.sidfmd_hv_categories.build(sidfmd_category_id: j[:id],
                                        name: j[:name])
      end
    end

    if modify then
      host.sidfmd_hv_created_ons.delete_all
      hv_created_ons = []
      row[:created_ons].each do |j|
        created_on = SidfmdHvCreatedOn.new
        created_on.sidfmd_host_view_id = host.id
        created_on.dt = j
        hv_created_ons << created_on
      end
      SidfmdHvCreatedOn.import(hv_created_ons, validate: false) if hv_created_ons.any?
    else
      row[:created_ons].each do |j|
        host.sidfmd_hv_created_ons.build(dt: j)
      end
    end

    if modify then
      host.sidfmd_hv_due_dates.delete_all
      hv_due_dates = []
      row[:due_dates].each do |j|
        due_date = SidfmdHvDueDate.new
        due_date.sidfmd_host_view_id = host.id
        due_date.d = j
        hv_due_dates << due_date
      end
      SidfmdHvDueDate.import(hv_due_dates, validate: false) if hv_due_dates.any?
    else
      row[:due_dates].each do |j|
        host.sidfmd_hv_due_dates.build(d: j)
      end
    end

    if modify then
      host.sidfmd_hv_unf_due_dates.delete_all
      hv_unf_due_dates = [] 
      row[:due_dates_unfinished].each do |j|
        due_date = SidfmdHvUnfDueDate.new
        due_date.sidfmd_host_view_id = host.id
        due_date.d = j
        hv_unf_due_dates << due_date
      end
      SidfmdHvUnfDueDate.import(hv_unf_due_dates, validate: false) if hv_unf_due_dates.any?
    else
      row[:due_dates_unfinished].each do |j|
        host.sidfmd_hv_unf_due_dates.build(d: j)
      end
    end

    if modify then
      host.sidfmd_hv_risks.delete_all
      hv_risks = []
      if row[:softek] and row[:softek][:risk_ids] then
        row[:softek][:risk_ids].each do |j|
          risk = SidfmdHvRisk.new
          risk.sidfmd_host_view_id = host.id
          risk.risk_id = j
          hv_risks << risk
        end
      end
      SidfmdHvRisk.import(hv_risks, validate: false) if hv_risks.any?
    else
      if row[:softek] and row[:softek][:risk_ids] then
        row[:softek][:risk_ids].each do |j|
          host.sidfmd_hv_risks.build(risk_id: j)
        end
      end
    end

    if modify then
      host.sidfmd_hv_areas.delete_all
      hv_areas = []
      if row[:softek] and row[:softek][:area_ids] then
        row[:softek][:area_ids].each do |j|
          area = SidfmdHvArea.new
          area.sidfmd_host_view_id = host.id
          area.area_id = j
          hv_areas << area
        end
      end
      SidfmdHvArea.import(hv_areas, validate: false) if hv_areas.any?
    else
      if row[:softek] and row[:softek][:area_ids] then
        row[:softek][:area_ids].each do |j|
          host.sidfmd_hv_areas.build(area_id: j)
        end
      end
    end
  end

  # SidfmdHost#create_all_contentのデータ1行の値を反映する。
  def apply_row(row)
    self.class.build_row(self, row, true)
  end

  def self.update_os(sidfmd_host_ids)
    modify_hostids = SidfmdHostView.where(:sidfmd_host_id => sidfmd_host_ids).map(&:sidfmd_host_id)
    if modify_hostids.any? then
      host_contents = SidfmdHost.create_all_content(no_cache: true, category_id: 0, role_admin: true, corresp_required: Sidfm.corresp_required, host_ids: modify_hostids)
      modify_hostids.each do |sidfmd_host_id|
        row = host_contents.find {|x| x[:sidfmd_host_id] == sidfmd_host_id }
        if row then
          SidfmdHostView.transaction do
            view = SidfmdHostView.lock.find_by_sidfmd_host_id(sidfmd_host_id)
            if view then
              if row[:os] and row[:os].size > 0 then
                view.osname = row[:os][:osname]
                if row[:os][:dist] then
                  view.queue_dist_id = row[:os][:dist].id
                elsif row[:os][:app] and row[:os][:app].size > 0 then
                  app = row[:os][:app]
                  app = row[:os][:app].first if app.kind_of?(Array)
                  if app[:product] then
                    view.product_id = app[:product].pid
                  end
                  if app[:version] then
                    view.version_id = app[:version].vid
                    unless view.product_id then
                      begin
                        view.product_id = app[:version].sidfmd_product.pid
                      rescue Exception
                      end
                    end
                  end
                end
              else
                view.product_id = nil
                view.version_id = nil
                view.queue_dist_id = nil
                view.osname = nil
              end  
              view.save 
            end
          end
        end
      end
    end
  end

end
