# -*- coding: utf-8 -*-

class SidfmdRule < ActiveRecord::Base
  belongs_to :sidfmd_category

  has_one :sidfmd_rule_trigger, dependent: :destroy
  has_many :sidfmd_rule_conds, dependent: :destroy
  has_many :sidfmd_rule_actions, dependent: :destroy

  scope :enables, -> { where(:flg_enable => true) }
  scope :filter_category, -> (category_id) { where(:sidfmd_category_id => category_id) }

  #キーワードの変換時とhtmlのタイトル表示に使用する
  PAGE_TITLE = "ルール設定"

  #ルールの最大登録数
  MAXIMUM_LIMIT = 500

  #triggerの一覧。パラメータに使用する。
  RULE_TRIGGER_NAMES = {
      'periodic_hour' => "毎日実行-時間指定",
      'periodic_interval'=> "定期実行-間隔指定",
      'vulsync' => "脆弱性同期時 実行",
  }

  #condの一覧。htmlの要素やパラメータに使用する。
  RULE_COND_NAMES = {
      'issue-update' => '脆弱性の更新',
      'cvss-exploitable' => '攻撃コード指定',
      'sri-above' => 'SRI指定値以上',
      'sri-below' => 'SRI指定値以下',
      'sri-warning' => 'SRI値要対応',
      'cvss-above' => 'CVSS指定値以上',
      'cvss-below' => 'CVSS指定値以下',
      'cvss-warning' => 'CVSS値要対応',
      'cvss-notreg' => 'CVSS値未定義',
      'cvss-vector' => 'CVSS指標',
      'status-match' => 'ステータスの一致',
      'product-match' => 'プロダクト限定',
      'product-unmatch' => 'プロダクト除外',
      'due-date-expired' => '対策予定日超過',
      'due-date-approach' => '対策予定日接近',
      'history-create' => '履歴追加',
      'category-match' => 'カテゴリ限定',
      'category-unmatch' => 'カテゴリ除外',
      'host-match' => 'ホスト限定',
      'host-unmatch' => 'ホスト除外',
      'host-syncing' => '同期中ホスト',
      'timerange' => '期間指定',
  }

  #action一覧。htmlの要素やパラメータに使用する。
  RULE_ACTION_NAMES = {
    'send-mail-user' => 'メール送信(ユーザ)',
    'send-mail-category'=> 'メール送信(カテゴリ)',
    'status-change' => 'ステータス変更',
    'due-date-change' => '対策予定日設定',
    'send-syslog' => 'syslog送信',
  }

  #操作のsyslog送信で指定可能なログの種類
  SYSLOG_FACILITIES = SyslogRuby::Facility.constants.find_all {|x| /^LOG/ !~ x.to_s }.map(&:to_s)

  #操作のsyslog送信で指定可能なログの深刻度
  SYSLOG_SEVERITIES = SyslogRuby::Severity.constants.find_all {|x| /^(LOG|WARNING)/ !~ x.to_s }.map(&:to_s)
  
  # ログでのTrigger, Cond, Actionの名称
  LOG_TRIGGER = 'trigger'.freeze
  LOG_COND = 'cond'.freeze
  LOG_ACTION = 'action'.freeze
  LOG_KINDS = {
    LOG_TRIGGER => 'タイミング'.freeze,
    LOG_COND => '条件'.freeze,
    LOG_ACTION => '操作'.freeze,
  }.freeze

  # ログの種類
  LOG_INFO = 'info'.freeze
  LOG_WARN = 'warn'.freeze
  LOG_ERR = 'err'.freeze
  LOG_TYPES = {
    LOG_INFO => '情報'.freeze,
    LOG_WARN => '警告'.freeze,
    LOG_ERR => 'エラー'.freeze,
  }.freeze

  # Trigger, Cond, Actionのそれぞれのlogsに保存できるログの最大件数
  # 1つのルールには1つのTriggerと複数のCond, Actionがあるので、合計するとそれなりの
  # 量になる。
  LOG_MAXNUM = 30

  # ログの追加処理
  # Trigger, Cond, Actionで利用
  def self.log_append(logs, type, kind, message)
    ret = nil
    begin
      ret = JSON.parse(logs)
    rescue JSON::ParserError
      ret = []
    end
    ret << { 'type' => type, 'kind' => kind, 'message' => message, 'created_at' => Time.current.iso8601(3) }
    ret.compact!
    ret.sort! {|x, y|
      (y['created_at'] && x['created_at']) ?
      y['created_at'] <=> x['created_at'] :
      (y['created_at'] ? 1 : -1)
    }
    JSON.generate(ret[0, LOG_MAXNUM])
  end

  # 指定した現在の条件に適合するルールのIDの配列を返す。
  def self.match_rule_ids(vulsync: false, periodic: false)
    ret = []
    # 定期実行
    if periodic then
      # hourは、前回実行時刻から20時間以上経過及び
      # 現在時刻がhour時になっていることが条件。
      # 始めての実行時(sidfmd_rule_triggers.last_checked_at=nil)は、後者の条件のみ。
      # current_timestampは、sidfmd_rule_triggers.last_checked_atがwithout time zone
      # だったため、整合性を採るために細工したもの。
      current_timestamp = "'#{Time.current.localtime.strftime('%Y/%m/%d %H:%M:%S')}'::timestamp without time zone"
      SidfmdRuleTpHour.enables
        .where("
          sidfmd_rule_tp_hours.hour IS NOT NULL AND (
            (
              sidfmd_rule_triggers.last_checked_at IS NULL AND
              TO_CHAR(#{current_timestamp}, 'HH24')::integer = sidfmd_rule_tp_hours.hour
            ) OR (
              sidfmd_rule_triggers.last_checked_at IS NOT NULL AND
              (#{current_timestamp} - sidfmd_rule_triggers.last_checked_at) >= '20 hours'::interval AND
              TO_CHAR(#{current_timestamp}, 'HH24')::integer = sidfmd_rule_tp_hours.hour
            )
          )
      ").each do |tph|
        ret << tph.sidfmd_rule_trigger.sidfmd_rule_id
      end
      # intervalは前回実行時間から、interval時間以上経過していることが条件
      # 初めての実行時(sidfmd_rule_triggers.last_checked_atがnil)は、無条件で適合。
      SidfmdRuleTpInterval.enables
        .where("
           (
             sidfmd_rule_tp_intervals.unit IS NOT NULL AND
             sidfmd_rule_tp_intervals.value IS NOT NULL ) AND (
             sidfmd_rule_triggers.last_checked_at IS NULL OR
             (#{current_timestamp} - sidfmd_rule_triggers.last_checked_at) >=
              (sidfmd_rule_tp_intervals.value::text || ' ' || sidfmd_rule_tp_intervals.unit || 's')::interval
           )
      ").each do |tpi|
        ret << tpi.sidfmd_rule_trigger.sidfmd_rule_id
      end
    end
    # 脆弱性同期時
    if vulsync then
      # vulsyncは無条件で適合。
      SidfmdRuleTVulsync.enables.each do |i|
        ret << i.sidfmd_rule_trigger.sidfmd_rule_id
      end
    end
    ### FIXME: 有効なruleのtriggerについて、logsに成否の結果を記録。
    ret
  end

  # ルールのTrigger, Cond, Actionのログをまとめて時系列でソートし返す。
  # 以下のHashの配列を返す。
  # {
  #   'type'       LOG_TYPES.keysのいずれか。LOG_TYPESで日本語名称を取得して出力すること。
  #   'kind'       LOG_KINDS.keysのいずれか。LOG_KINDSで日本語名称を朱徳して出力すること。
  #   'message'    ログ内容
  #   'created_at' 記録日時
  # }
  def logs(maxlogs = 200)
    ret = []
    if self.sidfmd_rule_trigger then
      begin
        r = JSON.parse(self.sidfmd_rule_trigger.logs)
        r.each do |i|
          ret << i.merge('kind' => SidfmdRule::LOG_TRIGGER)
        end
      rescue JSON::ParserError => e
        logger.error(e.message)
        e.backtrace.each do |i|
          logger.error(i)
        end
      end
    end
    self.sidfmd_rule_conds.each do |cond|
      begin
        r = JSON.parse(cond.logs)
        r.each do |i|
          ret << i.merge('kind' => SidfmdRule::LOG_COND)
        end
      rescue JSON::ParserError => e
        logger.error(e.message)
        e.backtrace.each do |i|
          logger.error(i)
        end
      end
    end
    self.sidfmd_rule_actions.each do |action|
      begin
        r = JSON.parse(action.logs)
        r.each do |i|
          ret << i.merge('kind' => SidfmdRule::LOG_ACTION)
        end
      rescue JSON::ParserError => e
        logger.error(e.message)
        e.backtrace.each do |i|
          logger.error(i)
        end
      end
    end
    ret.compact!
    ret.sort! {|x, y|
      (y['created_at'] && x['created_at']) ?
      y['created_at'] <=> x['created_at'] :
      (y['created_at'] ? 1 : -1)
    }
    # 時刻のフォーマットの変更
    ret.each do |i|
      if i['created_at'] then
        i['created_at'].gsub!(/^(\d{4}-\d{2}-\d{2})T(\d{2}:\d{2}:\d{2}\.\d+)[+-]\d{2}:\d{2}$/, '\1 \2')
      end
    end
    ret[0, maxlogs].reverse
  end

  #指定したユーザに関連しているルールの情報を返す。
  #sidfm_users_controller/sidfm_internal_api_controllerで使用
  #関係するのは今のところユーザ削除とカテゴリ変更のみ
  #actionは今のところimport/category_change/bulk_category_change/deleteのいずれか
  #:change_category_user_idsにはカテゴリ変更するユーザのidsを、:delete_user_idsには削除するユーザのidsを入れて渡す
  def self.users_related_rule_info(**opts)
    #初めの文字がアンダースコアの変数にしてあるものは再定義しないこと
    user_ids = []
    
    unless opts[:action] then
      raise ArgumentError, "must be action" unless opts.has_key?(:action)
    end

    case opts[:action]
    when 'import' then
      #インポート
      if !opts[:delete_user_ids] or !opts[:change_category_user_ids] then
         raise ArgumentError, "must be delete_user_ids and change_category_user_ids in case of the import"
      end
    when  'delete' then
      #削除
      if !opts[:delete_user_ids] then
        raise ArgumentError, "must be delete_user_ids in case of the delete" 
      end
    else
      #カテゴリ(一括)変更
      if !opts[:change_category_user_ids] then
        raise ArgumentError, "must be delete_user_ids in case of the category_change" 
      end
    end

    if opts[:delete_user_ids] then
      opts[:delete_user_ids] = opts[:delete_user_ids].map(&:to_i)
      user_ids |= opts[:delete_user_ids]
    end
    if opts[:change_category_user_ids] then
      user_ids |= opts[:change_category_user_ids].keys.map(&:to_i)
    end

    unless opts[:role_admin] then
      unless opts[:user_category_ids] then
        raise ArgumentError, "must be user_category_ids in case the user's role is not admin"
      end
    end

    _rules = if user_ids.any? then
      SidfmdRule.where(%|
        id IN (?) OR id IN (?)
      |,SidfmdRuleAction.joins(%|
          LEFT OUTER JOIN sidfmd_rule_a_mail_categories ON sidfmd_rule_a_mail_categories.sidfmd_rule_action_id = sidfmd_rule_actions.id
        |)
        .joins(%|
          LEFT OUTER JOIN sidfmd_rule_a_mail_users ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id
        |).where(%|
          sidfmd_rule_a_mail_categories.from_user_id IN (?) OR
          sidfmd_rule_a_mail_users.from_user_id IN (?) OR
          sidfmd_rule_a_mail_users.user_id IN (?)
        |, user_ids, user_ids, user_ids)
        .select(:sidfmd_rule_id),#ユーザに関連する操作
        SidfmdRuleCond.joins(%|
          LEFT OUTER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.sidfmd_rule_cond_id = sidfmd_rule_conds.id
        |).joins(%|
          LEFT OUTER JOIN sidfmd_rule_chp_user_subs ON sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id = sidfmd_rule_c_history_posts.id
        |)
        .where(:sidfmd_rule_chp_user_subs => { :user_id => user_ids })
        .select(:sidfmd_rule_id) #ユーザに関連する条件
      )
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_a_mail_users
          INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id
          WHERE sidfmd_rule_id = sidfmd_rules.id
        ) AS count_to_user_ids /*SidfmdRuleAMailUserの送信先に設定されているユーザの数*/
      ")
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_actions
          WHERE sidfmd_rules.id = sidfmd_rule_id
        ) AS count_actions /*ルールのactionの数*/      
      ")
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_chp_user_subs
          INNER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.id = sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id
          INNER JOIN sidfmd_rule_conds ON sidfmd_rule_conds.id = sidfmd_rule_c_history_posts.sidfmd_rule_cond_id
          WHERE sidfmd_rule_id = sidfmd_rules.id
        ) AS count_history_cond_user_ids /*SidfmdRuleChpUserSubの数(SidfmdRuleCHistoryPostに設定されているユーザの数)*/
      ")
      .select("
        (
          SELECT COUNT(*)
          FROM sidfmd_rule_conds
          WHERE sidfmd_rules.id = sidfmd_rule_id
        ) AS count_conds /*ルールの操作の数*/
      ")
      .select(:id, :name, :sidfmd_category_id)
    else
      SidfmdRule.where(:id => 0)
    end

    if opts[:delete_user_ids] then
      opts[:delete_user_ids] = [ 0 ] if opts[:delete_user_ids].size == 0
      _rules = _rules.select(" 
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.user_id IN (SELECT unnest(ARRAY[#{opts[:delete_user_ids].join(',')}]))
          )
        ) AS orig_delete_mail_users_to_actions /*SidfmdRuleAMailUserの送信先に設定されている削除対象のユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.from_user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.from_user_id IN (SELECT unnest(ARRAY[#{opts[:delete_user_ids].join(',')}]))
          )
        ) AS orig_delete_mail_users_from_actions /*SidfmdRuleAMailUserの送信元に設定されている削除対象のユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_categories.from_user_id], ':')
            FROM sidfmd_rule_a_mail_categories
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_categories.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_categories.from_user_id IN (SELECT unnest(ARRAY[#{opts[:delete_user_ids].join(',')}]))
          )
        ) AS orig_delete_mail_categories_from_actions /*SidfmdRuleAMailCategoryの送信元に設定されている削除対象のユーザに関連するactions(通常は1つのみ)*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_chp_user_subs.id, sidfmd_rule_chp_user_subs.user_id, sidfmd_rule_conds.id], ':')
            FROM sidfmd_rule_chp_user_subs
            INNER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.id = sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id
            INNER JOIN sidfmd_rule_conds ON sidfmd_rule_conds.id = sidfmd_rule_c_history_posts.sidfmd_rule_cond_id
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_chp_user_subs.user_id IN (SELECT unnest(ARRAY[#{opts[:delete_user_ids].join(',')}]))
          )
        ) AS orig_delete_history_cond_chp_users /*SidfmdRuleChpUserSubのidとuser_idとsidfmd_rule_cond_idのJsonObject*/
      ") 
    end
    if opts[:change_category_user_ids] then
      target_user_ids = opts[:change_category_user_ids].keys.map(&:to_i)
      target_user_ids = [ 0 ] if target_user_ids.size == 0
      _rules = _rules.select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_category_mail_users_to_actions /*SidfmdRuleAMailUserの送信先に設定されているカテゴリ変更ユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_users.from_user_id], ':')
            FROM sidfmd_rule_a_mail_users 
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_users.from_user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_category_mail_users_from_actions /*SidfmdRuleAMailUserの送信元に設定されているカテゴリ変更ユーザに関連するactions*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_actions.id, sidfmd_rule_a_mail_categories.from_user_id], ':')
            FROM sidfmd_rule_a_mail_categories
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_categories.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_a_mail_categories.from_user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_category_mail_categories_from_actions /*SidfmdRuleAMailCategoryの送信元に設定されているカテゴリ変更ユーザに関連するactions(通常は1つのみ)*/
      ")
      .select("
        (
          ARRAY(
            SELECT ARRAY_TO_STRING(ARRAY[sidfmd_rule_chp_user_subs.id, sidfmd_rule_chp_user_subs.user_id, sidfmd_rule_conds.id], ':')
            FROM sidfmd_rule_chp_user_subs
            INNER JOIN sidfmd_rule_c_history_posts ON sidfmd_rule_c_history_posts.id = sidfmd_rule_chp_user_subs.sidfmd_rule_c_history_post_id
            INNER JOIN sidfmd_rule_conds ON sidfmd_rule_conds.id = sidfmd_rule_c_history_posts.sidfmd_rule_cond_id
            WHERE sidfmd_rule_id = sidfmd_rules.id AND sidfmd_rule_chp_user_subs.user_id IN (SELECT unnest(ARRAY[#{target_user_ids.join(',')}]))
          )
        ) AS orig_change_category_history_cond_chp_users /*SidfmdRuleChpUserSubのidとuser_idとsidfmd_rule_cond_idのJsonObject*/
      ") 
    end

    _mail_cat_action_name = RULE_ACTION_NAMES['send-mail-category'];
    _mail_user_action_name = RULE_ACTION_NAMES['send-mail-user'];
    _history_cond_name = RULE_COND_NAMES['history-create']; 
    _rule_messages = {}
    _delete_rules = []
    _delete_actions = []
    _delete_conds = []
    _delete_chp_user_subs = []

    get_user_names = ->(objects){
       User.where(
         :id => objects.map{|action| action['user_id']}
       ).map{|user| "#{user.lastname} #{user.firstname}(ログイン名: #{user.login})"}
    }

    classify_action = ->(
      action, 
      action_name, 
      delete_mail_to_users_actions, 
      delete_mail_from_users_actions , 
      delete_mail_category_actions,
      user_category_ids,
      rule_category_id
    ){
      user_cat_ids = user_category_ids[action['user_id'].to_s]
      if user_cat_ids and (user_cat_ids.size == 0 or (rule_category_id and !user_cat_ids.include?(rule_category_id))) then
        if action_name == 'mail_to_users' then
          delete_mail_to_users_actions << action
        elsif action_name == 'mail_from_users' then
          delete_mail_from_users_actions << action
        else
          delete_mail_category_actions << action
        end
      end
    }

    classify_chp_user = ->(history_cond_chp_user, rule_delete_chp_user_subs, user_category_ids, rule_category_id){
      user_cat_ids = user_category_ids[history_cond_chp_user["user_id"].to_s]
      if user_cat_ids and (user_cat_ids.size == 0 or (rule_category_id and !user_cat_ids.include?(rule_category_id))) then
        rule_delete_chp_user_subs << history_cond_chp_user
      end
    }

    #ruleのObjectの生成とメソッドの追加
    create_rule_object = ->(rule){
      if rule.respond_to?(:orig_delete_mail_users_to_actions) then
        def rule.delete_mail_users_to_actions
          self.orig_delete_mail_users_to_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_delete_mail_users_from_actions) then
        def rule.delete_mail_users_from_actions
          self.orig_delete_mail_users_from_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end
   
      if rule.respond_to?(:orig_delete_mail_categories_from_actions) then
        def rule.delete_mail_categories_from_actions
          self.orig_delete_mail_categories_from_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_delete_history_cond_chp_users) then
        def rule.delete_history_cond_chp_users
          self.orig_delete_history_cond_chp_users.map{|cond|
            cols = cond.split(':')
            {
              'chp_user_sub_id' => cols[0].to_i,
              'user_id' => cols[1].to_i,
              'cond_id' => cols[2].to_i,
            }
          }
        end
      end

      if rule.respond_to?(:orig_change_category_mail_users_to_actions) then
        def rule.change_category_mail_users_to_actions
          self.orig_change_category_mail_users_to_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_change_category_mail_users_from_actions) then
        def rule.change_category_mail_users_from_actions
          self.orig_change_category_mail_users_from_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end

      if rule.respond_to?(:orig_change_category_mail_categories_from_actions) then
        def rule.change_category_mail_categories_from_actions
          self.orig_change_category_mail_categories_from_actions.map{|action|
            actionid_userid = action.split(':')
            { 'action_id' => actionid_userid[0].to_i, 'user_id' => actionid_userid[1].to_i }
          }
        end
      end
     
      if rule.respond_to?(:orig_change_category_history_cond_chp_users) then
        def rule.change_category_history_cond_chp_users
          self.orig_change_category_history_cond_chp_users.map{|cond|
            cols = cond.split(':')
            {
              'chp_user_sub_id' => cols[0].to_i,
              'user_id' => cols[1].to_i,
              'cond_id' => cols[2].to_i,
            }
          }
        end
      end

      rule
    }

    case opts[:action]
    when 'delete' then 
      #ユーザ削除の場合
      _rules.each do |rule|
        rule = create_rule_object.(rule)
        #条件が全てなくなるかのチェック
        rule_delete_conds = []
        if rule.orig_delete_history_cond_chp_users.size > 0 and rule.orig_delete_history_cond_chp_users.size == rule.count_history_cond_user_ids then
          #履歴追加の条件に設定されているユーザが全て削除されるなら条件削除
          rule_delete_conds |= rule.delete_history_cond_chp_users
        end
        if rule.count_conds == rule_delete_conds.uniq{|cond| cond['cond_id']}.size then
          #条件がなくなるならルールを削除
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_conds }
          if opts[:messages_necessary] then
            reason = "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザが削除されるため。"
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => reason
            }
          end
        end

        #操作が全てなくなるかのチェック
        rule_delete_actions = []
        rule_delete_actions |= rule.delete_mail_users_to_actions
        rule_delete_actions |= rule.delete_mail_users_from_actions
        rule_delete_actions |= rule.delete_mail_categories_from_actions
        if rule.count_actions == rule_delete_actions.uniq{|action| action['action_id']}.size then
          #操作が無くなるルールは削除
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_actions }
          if opts[:messages_necessary] then
            reason = if rule.count_actions == rule.orig_delete_mail_users_from_actions.size then
                       "ルールの操作が「#{_mail_user_action_name}」のみで、送信元に設定されているユーザが削除されるため。"
                     elsif rule.count_actions == rule.orig_delete_mail_users_to_actions.size then
                       "ルールの操作が「#{_mail_user_action_name}」のみで、送信先に設定されている全てのユーザが削除されるため。"
                     elsif rule.count_actions == rule.orig_delete_mail_categories_from_actions.size then
                       "ルールの操作が「#{_mail_cat_action_name}」のみで、送信元に設定されているユーザが削除されるため。"
                     else
                       "ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、このユーザを削除すると有効な操作がなくなるため。"
                     end
            _rule_messages[rule.id] = {
              :rule_name => rule.name, 
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}#{reason}"
            }
          end
        end
        if _rule_messages[rule.id] then
          next
        end
  
        #条件のチェック
        _delete_conds |= rule_delete_conds
        if rule.orig_delete_history_cond_chp_users.size > 0 then
          if rule_delete_conds.find{|cond| cond['cond_id'] == rule.delete_history_cond_chp_users.first['cond_id']} then
            if opts[:messages_necessary] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "条件",
                :content => "「#{_history_cond_name}」が削除されます。",
                :reason => "「#{_history_cond_name}」に設定されている全てのユーザが削除されるため。"
              }
            end
          else
            rule_delete_chp_user_subs = rule.delete_history_cond_chp_users
            _delete_chp_user_subs |= rule_delete_chp_user_subs
            if opts[:messages_necessary] then
              user_names = get_user_names.(rule_delete_chp_user_subs)
              if user_names.any? then
                _rule_messages[rule.id] = {
                  :rule_name => rule.name,
                  :item_name => "条件",
                  :content => "「#{_history_cond_name}」から、#{user_names.join(', ')}が削除されます。",
                  :reason => "ユーザが削除されるため。"
                }
              end
            end
          end
        end
  
        #操作のチェック
        _delete_actions |= rule_delete_actions
        if opts[:messages_necessary] then
          content_msgs = []
          reason_msgs = []
          if rule.orig_delete_mail_categories_from_actions.any? then
            #メール送信(カテゴリ)の送信先ユーザが削除されるならメール送信(カテゴリ)が削除されるメッセージを表示
            content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_cat_action_name}」の送信元に設定されているユーザが削除されるため。"
          end
          if rule.orig_delete_mail_users_from_actions.any? then
            #メール送信(ユーザ)の送信元ユーザが削除されるならメール送信(ユーザ)が削除されるメッセージを表示
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_user_action_name}」の送信元に設定されているユーザが削除されるため。"
          elsif rule.orig_delete_mail_users_to_actions.any? then
            if rule.orig_delete_mail_users_to_actions.size == rule.count_to_user_ids then
              #メール送信(ユーザ)の送信先ユーザが全て削除されるならメール送信(ユーザ)が削除されるメッセージを表示
              content_msgs << "「#{_mail_user_action_name}」が削除されます。"
              reason_msgs <<  "「#{_mail_user_action_name}」の送信先に設定されているユーザが全て削除されるため。"
            else
              #メール送信(ユーザ)の送信先ユーザが削除されるメッセージを表示
              user_names = get_user_names.(rule.delete_mail_users_to_actions) 
              if user_names.any? then
                content_msgs << "「#{_mail_user_action_name}」の送信先から、#{user_names.join(', ')}が削除されます。"
                reason_msgs <<  "ユーザが削除されるため。"
              end
            end
          end
          if content_msgs.any? then
            if _rule_messages[rule.id] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "条件・操作",
                :content => _rule_messages[rule.id][:content] + "\n" + content_msgs.join("\n"),
                :reason => _rule_messages[rule.id][:reason] + "\n" + reason_msgs.join("\n")
              }
            else
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "操作",
                :content => content_msgs.join("\n"),
                :reason => reason_msgs.join("\n")
              }
            end
          end
        end
      end
    when 'category_change', 'bulk_category_change' then
      user_category_ids = opts[:change_category_user_ids] #opts[:change_category_user_ids]のkeyは文字列のユーザid, valueは変更するカテゴリのids
      _rules.each do |rule|
        rule = create_rule_object.(rule)
        rule_delete_conds = []
        #条件がなくなるかのチェック
        rule_delete_chp_user_subs = []
        
        rule.change_category_history_cond_chp_users.each do |chp_user|
          classify_chp_user.(chp_user, rule_delete_chp_user_subs, user_category_ids, rule.sidfmd_category_id)
        end
        if rule_delete_chp_user_subs.size > 0 then
          if rule_delete_chp_user_subs.size == rule.count_history_cond_user_ids then
            #全てのユーザのカテゴリがルールのカテゴリから外れるなら条件削除
            rule_delete_conds |= rule.change_category_history_cond_chp_users
          end   
        end

        if rule.count_conds == rule_delete_conds.uniq{|cond| cond['cond_id']}.size then
          #全ての条件がなくなるならルール削除
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_conds }
          if opts[:messages_necessary] then
            reason = "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => reason
            }
          end
        end

        #操作が無くなるかのチェック
        delete_mail_to_users_actions = []
        delete_mail_from_users_actions = []
        delete_mail_category_actions = []
        rule.change_category_mail_categories_from_actions.each do |action|
          classify_action.(
            action,
            'mail_category',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_category_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_category_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        rule.change_category_mail_users_from_actions.each do |action|
          classify_action.(
            action,
            'mail_from_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_from_users_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_from_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_from_users_actions.size + delete_mail_category_actions.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_from_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        rule.change_category_mail_users_to_actions.each do |action|
          classify_action.(
            action,
            'mail_to_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_to_users_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_to_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_to_users_actions.size + delete_mail_category_actions.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_to_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元と「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        if _rule_messages[rule.id] then
          next
        end

        _delete_conds |= rule_delete_conds
        #条件のチェック
        if rule_delete_chp_user_subs.size > 0 then
          if rule_delete_conds.find{|cond| cond['cond_id'] == rule.change_category_history_cond_chp_users.first['cond_id']} then
            if opts[:messages_necessary] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "条件",
                :content => "「#{_history_cond_name}」が削除されます。",
                :reason => "「#{_history_cond_name}」に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
              }
            end
          else
            _delete_chp_user_subs |= rule_delete_chp_user_subs
            if opts[:messages_necessary] then
              user_names = get_user_names.(rule_delete_chp_user_subs)
              if user_names.any? then
                _rule_messages[rule.id] = {
                  :rule_name => rule.name,
                  :item_name => "条件",
                  :content => "「#{_history_cond_name}」から、#{user_names.join(', ')}が削除されます。",
                  :reason => "ユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                }
              end
            end
          end
        end

        #操作のチェック
        _delete_actions |= (delete_mail_category_actions + delete_mail_to_users_actions + delete_mail_from_users_actions)
        
        if opts[:messages_necessary] then
          content_msgs = []
          reason_msgs = []
          if delete_mail_category_actions.any? then
            content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          end

          if delete_mail_from_users_actions.any? then
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          elsif delete_mail_to_users_actions.any? then
            if delete_mail_to_users_actions.size == rule.count_to_user_ids then
              content_msgs << "「#{_mail_user_action_name}」が削除されます。"
              reason_msgs << "「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            else
              user_names = get_user_names.(delete_mail_to_users_actions)
              if user_names.any? then
                content_msgs << "「#{_mail_user_action_name}」の送信先から、#{user_names.join(', ')}が削除されます。"
                reason_msgs <<  "ユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
              end
            end
          end

          if content_msgs.any? then
            if _rule_messages[rule.id] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "条件・操作",
                :content => _rule_messages[rule.id][:content] + "\n" + content_msgs.join("\n"),
                :reason => _rule_messages[rule.id][:reason] + "\n" + reason_msgs.join("\n")
              }
            else
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "操作",
                :content => content_msgs.join("\n"),
                :reason => reason_msgs.join("\n")
              }
            end
          end
        end
      end
    when 'import' then
      user_category_ids = opts[:change_category_user_ids] #opts[:change_category_user_ids]のkeyは文字列のユーザid, valueは変更するカテゴリのids
      _rules.each do |rule|
        rule = create_rule_object.(rule)
        rule_delete_conds = []
        #条件が全てなくなるかのチェック
        if rule.orig_delete_history_cond_chp_users.size > 0 and rule.orig_delete_history_cond_chp_users.size == rule.count_history_cond_user_ids then
          #履歴追加の条件に設定されているユーザが全て削除されるなら条件削除
          rule_delete_conds |= rule.delete_history_cond_chp_users
        end
        if rule.count_conds == rule_delete_conds.uniq{|cond| cond['cond_id']}.size then
          #条件がなくなるならルールを削除
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_conds }
          if opts[:messages_necessary] then
            reason = "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザが削除されるため。"
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => reason
            }
          end
        end

        #操作が全てなくなるかのチェック
        delete_user_actions = []
        #ユーザ削除によって操作が無くなるルールは削除
        delete_user_actions |= rule.delete_mail_users_to_actions
        delete_user_actions |= rule.delete_mail_users_from_actions
        delete_user_actions |= rule.delete_mail_categories_from_actions
        if rule.count_actions == delete_user_actions.uniq{|action| action['action_id']}.size then
          _delete_rules << { :rule_id => rule.id, :opts => delete_user_actions }
          if opts[:messages_necessary] then
            reason = if rule.count_actions == rule.orig_delete_mail_users_from_actions.size then
                       "ルールの操作が「#{_mail_user_action_name}」のみで、送信元に設定されているユーザが削除されるため。"
                     elsif rule.count_actions == rule.orig_delete_mail_users_to_actions.size then
                       "ルールの操作が「#{_mail_user_action_name}」のみで、送信先に設定されている全てのユーザが削除されるため。"
                     elsif rule.count_actions == rule.orig_delete_mail_categories_from_actions.size then
                       "ルールの操作が「#{_mail_cat_action_name}」のみで、送信元に設定されているユーザが削除されるため。"
                     else
                       "ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、関連するユーザが全て削除されるため。"
                     end
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}#{reason}"
            }
          end
          next
        end

        if _rule_messages[rule.id] then
          next
        end

        #カテゴリ変更チェック
        #条件が全てなくなるかのチェック
        rule_delete_chp_user_subs = [] #カテゴリ変更によって削除する履歴追加の条件のchp_user_sub_ids
        rule.change_category_history_cond_chp_users.each do |chp_user|
          classify_chp_user.(chp_user, rule_delete_chp_user_subs, user_category_ids, rule.sidfmd_category_id)
        end
        if rule_delete_chp_user_subs.size > 0 then
          if (rule.orig_delete_history_cond_chp_users.size + rule_delete_chp_user_subs.size) == rule.count_history_cond_user_ids then
            #全てのユーザのカテゴリがルールのカテゴリから外れる、またはは削除されるなら条件削除
            rule_delete_conds |= (rule.change_category_history_cond_chp_users + rule.delete_history_cond_chp_users)
          end
        end

        if rule.count_conds == rule_delete_conds.uniq{|cond| cond['cond_id']}.size then
          #全ての条件がなくなるならルール削除
          _delete_rules << { :rule_id => rule.id, :opts => rule_delete_conds }
          if opts[:messages_necessary] then
            reason = if rule.orig_delete_history_cond_chp_users.size == rule.count_history_cond_user_ids then
                       "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザが削除されるため。"
                     elsif rule_delete_chp_user_subs.size == rule.count_history_cond_user_ids then
                       "ルールの条件が「#{_history_cond_name}」のみで、この条件に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                     else
                       "ユーザの削除及びカテゴリ変更によって、有効な条件が無くなるため。"
                     end
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => reason
            }
          end
        end

        delete_mail_to_users_actions = [] #カテゴリ変更によって削除するメール送信(ユーザ)の送信先のactionのids
        delete_mail_from_users_actions = [] #カテゴリ変更によって削除するメール送信(ユーザ)の送信元のactionのids      
        delete_mail_category_actions = [] #カテゴリ変更によって削除するメール送信(カテゴリ)のactionのids
        
        rule.change_category_mail_categories_from_actions.each do |action|
          classify_action.(
            action,
            'mail_category',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_category_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_category_actions }
          _rule_messages[rule.id] = {
            :rule_name => rule.name,
            :item_name => "ルール",
            :content => "ルールが削除されます。",
            :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          }
          next
        end

        rule.change_category_mail_users_from_actions.each do |action|
          classify_action.(
            action,
            'mail_from_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_from_users_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_from_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_from_users_actions.size + delete_mail_category_actions.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_from_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        rule.change_category_mail_users_to_actions.each do |action|
          classify_action.(
            action,
            'mail_to_users',
            delete_mail_to_users_actions,
            delete_mail_from_users_actions ,
            delete_mail_category_actions,
            user_category_ids,
            rule.sidfmd_category_id
          )
        end

        if delete_mail_to_users_actions.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => delete_mail_to_users_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」のみで、「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        elsif (delete_mail_to_users_actions.size + delete_mail_category_actions.size) == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => (delete_mail_to_users_actions + delete_mail_category_actions) }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ルールの操作が「#{_mail_user_action_name}」と「#{_mail_cat_action_name}」のみで、「#{_mail_cat_action_name}」の送信元と「#{_mail_user_action_name}」の送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
            }
          end
          next
        end

        merged_delete_mail_to_user_actions = (rule.delete_mail_users_to_actions + delete_mail_to_users_actions).uniq
        all_rule_delete_actions = (merged_delete_mail_to_user_actions + delete_user_actions + delete_mail_category_actions + delete_mail_from_users_actions).uniq
        #カテゴリ変更とユーザ削除によって操作が無くなるルールを削除
        if all_rule_delete_actions.uniq{|action| action['action_id']}.size == rule.count_actions then
          _delete_rules << { :rule_id => rule.id, :opts => all_rule_delete_actions }
          if opts[:messages_necessary] then
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "ルール",
              :content => "ルールが削除されます。",
              :reason => "#{_rule_messages[rule.id] ? "#{_rule_messages[rule.id][:reason]}また、" : ""}ユーザの削除及びカテゴリ変更によって、有効な操作がなくなるため。"
            }
          end
          next
        end

        if _rule_messages[rule.id] then
          next
        end

        _delete_conds |= rule_delete_conds
        #条件のチェック
        is_history_cond_deleted_by_user_delete = (rule.orig_delete_history_cond_chp_users.size > 0 and (rule_delete_conds.find{|cond| cond['cond_id'] == rule.delete_history_cond_chp_users.first['cond_id']} != nil))
        is_history_cond_deleted_by_user_cat_change = (rule.orig_change_category_history_cond_chp_users.size > 0 and (rule_delete_conds.find{|cond| cond['cond_id'] == rule.change_category_history_cond_chp_users.first['cond_id']} != nil))  
        
        if is_history_cond_deleted_by_user_delete or is_history_cond_deleted_by_user_cat_change then
          if opts[:messages_necessary] then
            content_message = if is_history_cond_deleted_by_user_delete and is_history_cond_deleted_by_user_cat_change then
                                "削除及びカテゴリ変更によって、「#{_history_cond_name}」に設定されている全てのユーザが無効になるため。"
                              elsif is_history_cond_deleted_by_user_delete then
                                "「#{_history_cond_name}」に設定されている全てのユーザが削除されるため。"
                              else
                                "「#{_history_cond_name}」に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                              end
            _rule_messages[rule.id] = {
              :rule_name => rule.name,
              :item_name => "条件",
              :content => "「#{_history_cond_name}」が削除されます。",
              :reason => "#{content_message}"
            } 
          end
        else
          merged_rule_delete_chp_user_subs = (rule.delete_history_cond_chp_users + rule_delete_chp_user_subs) 
          if merged_rule_delete_chp_user_subs.size > 0 then
            _delete_chp_user_subs |= merged_rule_delete_chp_user_subs
            if opts[:messages_necessary] then 
              user_names = get_user_names.(merged_rule_delete_chp_user_subs)
              if user_names.size > 0 then
                content_message = if rule.orig_delete_history_cond_chp_users.size > 0 and rule_delete_chp_user_subs.size > 0 then
                                    "ユーザが削除される、またはユーザのカテゴリがルールのカテゴリと一致しなくなるため。"
                                  elsif rule.orig_delete_history_cond_chp_users.size > 0 then
                                    "ユーザが削除されるため。"
                                  else
                                    "ユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                                  end
                _rule_messages[rule.id] = {
                  :rule_name => rule.name,
                  :item_name => "条件",
                  :content => "「#{_history_cond_name}」から、#{user_names.join(', ')}が削除されます。",
                  :reason => "#{content_message}"
                }
              end
            end
          end
        end

        #操作のチェック
        _delete_actions |= all_rule_delete_actions

        if opts[:messages_necessary] then
          content_msgs = []
          reason_msgs = []

          if rule.orig_delete_mail_categories_from_actions.any? then
            content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_cat_action_name}」の送信元に設定されているユーザが削除されるため。"
          elsif delete_mail_category_actions.any? then
            content_msgs << "「#{_mail_cat_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_cat_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          end

          if rule.orig_delete_mail_users_from_actions.any? then
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_user_action_name}」の送信元に設定されているユーザが削除されるため。"
          elsif delete_mail_from_users_actions.any? then
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << "「#{_mail_user_action_name}」の送信元に設定されているユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
          elsif rule.count_to_user_ids > 0 and merged_delete_mail_to_user_actions.size == rule.count_to_user_ids then
            #送信先ユーザがなくなるなら、メール送信(ユーザ)の操作を削除
            reason = if rule.count_to_user_ids == rule.orig_delete_mail_users_to_actions.size then
                       "送信先に設定されている全てのユーザが削除されるため。"
                     elsif rule.count_to_user_ids == delete_mail_to_users_actions.size then
                       "送信先に設定されている全てのユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                     else
                       "送信先に設定されている全てのユーザが無効になるため。"
                     end
            content_msgs << "「#{_mail_user_action_name}」が削除されます。"
            reason_msgs << reason
          elsif merged_delete_mail_to_user_actions.any? then
            user_names = get_user_names.(rule.delete_mail_users_to_actions + delete_mail_to_users_actions)
            if user_names.any? then
              content_msgs << "「#{_mail_user_action_name}」の送信先から、#{user_names.join(', ')}が削除されます。"
              reason_msgs << if rule.orig_delete_mail_users_to_actions.any? and delete_mail_to_users_actions.any? then
                               "ユーザが削除される、またはユーザのカテゴリがルールのカテゴリと一致しなくなるため。"
                             elsif rule.orig_delete_mail_users_to_actions.any? then
                               "ユーザが削除されるため。"
                             else
                               "ユーザのカテゴリが、ルールのカテゴリと一致しなくなるため。"
                             end
            end
          end
          if content_msgs.any? then
            if _rule_messages[rule.id] then
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "条件・操作",
                :content => _rule_messages[rule.id][:content] + "\n" + content_msgs.join("\n"),
                :reason => _rule_messages[rule.id][:reason] + "\n" + reason_msgs.join("\n")
              }
            else
              _rule_messages[rule.id] = {
                :rule_name => rule.name,
                :item_name => "操作",
                :content => content_msgs.join("\n"),
                :reason => reason_msgs.join("\n")
              }
            end
          end
        end
      end
    end

    delete_chp_user_sub_ids = _delete_chp_user_subs.map{|chp_user| chp_user['chp_user_sub_id']}.uniq
    unchangeable_users = []
    is_unchangeable_user = false

    if opts[:role_admin] then
      delete_rule_ids = _delete_rules.map{|rule| rule[:rule_id]}.uniq
      delete_action_ids = _delete_actions.map{|action| action['action_id']}.uniq
      delete_cond_ids = _delete_conds.map{|cond| cond['cond_id']}.uniq
    else
      #管理者で無ければ、カテゴリ範囲外のルールのルール自体が削除される、または操作・条件自体が削除される場合は変更できないユーザを取得する
      unchangeable_user_ids = []
      t_rules = {}
      _delete_rules.each do |rule|
        if t_rules[rule[:rule_id]] then
          t_rules[rule[:rule_id]] += rule[:opts]
        else
          t_rules[rule[:rule_id]] = rule[:opts]
        end
      end
      t_actions = {}
      t_action_ids = []
      _delete_actions.each do |action|
        t_actions[action['action_id']] = [] unless t_actions[action['action_id']]
        t_actions[action['action_id']] << action['user_id']
        t_action_ids << action['action_id']
      end
      t_conds = {}
      _delete_conds.each do |cond|
        t_conds[cond['cond_id']] = [] unless t_conds[cond['cond_id']]
        t_conds[cond['cond_id']] << cond['user_id']
      end
      
      unchangeable_rules = (opts[:user_category_ids].size > 0 ? SidfmdRule.where(%|sidfmd_category_id NOT IN (?) OR sidfmd_category_id IS NULL|, opts[:user_category_ids]) : SidfmdRule.all)
      unchangeable_rules.where(:id => t_rules.keys).each do |rule|
        if t_rules[rule.id] then
          #ルール自体が削除されるなら、ユーザの変更はできない
          unchangeable_user_ids |= t_rules[rule.id].map{|opt| opt['user_id']}
          t_rules.delete(rule.id)
          _rule_messages.delete(rule.id) if opts[:messages_necessary]
        end
      end

      unchangeable_rules.where(
        :id => SidfmdRuleAction.where(:id => t_action_ids).select(:sidfmd_rule_id) 
      )
      .select("
        (
          ARRAY(
            SELECT sidfmd_rule_actions.id 
            FROM sidfmd_rule_a_mail_users
            INNER JOIN sidfmd_rule_actions ON sidfmd_rule_a_mail_users.sidfmd_rule_action_id = sidfmd_rule_actions.id 
            WHERE sidfmd_rule_id = sidfmd_rules.id
          )
        ) AS mail_to_action_ids
      ")
      .select(:id)
      .each do |rule|
        if (rule.mail_to_action_ids - t_action_ids).size == 0 then
          #操作自体が削除されるなら、ユーザを変更できない
          rule.mail_to_action_ids.each do |action_id|
            unchangeable_user_ids |= t_actions[action_id] if t_actions[action_id]
            t_actions.delete(action_id)
          end
          _rule_messages.delete(rule.id) if opts[:messages_necessary]
        end
      end
      
      SidfmdRuleAction.joins(:sidfmd_rule_a_mail_category).where(%|
        sidfmd_rule_actions.id IN (?) AND sidfmd_rule_actions.sidfmd_rule_id IN (?)
      |,t_action_ids, unchangeable_rules.select(:id)
      )
      .select('sidfmd_rule_actions.id AS action_id', 'sidfmd_rule_actions.sidfmd_rule_id AS rule_id')
      .each do |action|
        unchangeable_user_ids |= t_actions[action.action_id] if t_actions[action.action_id]
        t_actions.delete(action.action_id)
        _rule_messages.delete(action.rule_id) if opts[:messages_necessary] 
      end
 
      SidfmdRuleCond.where(%|
        id IN (?) AND sidfmd_rule_id IN (?)
      |, t_conds.keys, unchangeable_rules.select(:id)
      )
      .select('id AS cond_id', 'sidfmd_rule_id AS rule_id')
      .each do |cond|
        unchangeable_user_ids |= t_conds[cond.cond_id] if t_conds[cond.cond_id]
        t_conds.delete(cond.cond_id)
        _rule_messages.delete(cond.rule_id) if opts[:messages_necessary]  
      end
      delete_rule_ids = t_rules.keys
      delete_action_ids = t_actions.keys
      delete_cond_ids = t_conds.keys
      if unchangeable_user_ids.any? then
        is_unchangeable_user = true
        if opts[:messages_necessary] then
          unchangeable_users = User.where(
            :id => unchangeable_user_ids
          ).map{|user|
            {
              :login => user.login,
              :lastname => user.lastname, 
              :firstname => user.firstname,
              :mail => user.mail
            }
         }
        end
      end
      if opts[:messages_necessary] then
        unchangeable_rules.where(:id => _rule_messages.keys).each do |rule|
          _rule_messages.delete(rule.id)
        end
      end
    end 

    ret = {
      :delete_rule_ids => delete_rule_ids,
      :delete_action_ids => delete_action_ids,
      :delete_cond_ids => delete_cond_ids,
      :delete_chp_user_sub_ids => delete_chp_user_sub_ids,
      :is_unchangeable_user => is_unchangeable_user,
      :unchangeable_users => unchangeable_users
    }

    ret[:rule_messages] = _rule_messages if opts[:messages_necessary]
    ret


  end

  #ホストに関連するルールを返す。
  #今は報告者権限で、ログインユーザのカテゴリ範囲外のルールの条件・操作・ルール自体が
  #削除される時に処理させないようにする機能のみ。
  #今後、sidfm_internal_api_cotrollerの処理をこちらに移動して、全面的に処理を修正する。
  def self.hosts_related_rule_info(**opts)
    unless opts[:role_admin] then
      unless opts[:user_category_ids] then
        raise ArgumentError, "must be user_category_ids in case the user's role is not admin"
      end
    end

    is_unchangeable_host = false
    unchangeable_hosts = []
    unless opts[:role_admin] then
      unchangeable_rules = (opts[:user_category_ids].size > 0 ? SidfmdRule.where(%|sidfmd_category_id NOT IN (?) OR sidfmd_category_id IS NULL|, opts[:user_category_ids]) : SidfmdRule.all)
      unchangeable_delete_rule_ids = unchangeable_rules.where(:id => opts[:delete_rule_ids]).pluck(:id)
      unchangeable_delete_cond_ids = SidfmdRuleCond.where(%|
        id IN (?) AND sidfmd_rule_id IN (?)
      |, opts[:delete_cond_ids], unchangeable_rules.select(:id)
      ).pluck(:id)
      #opts[:delete_rule_ids] = (opts[:delete_rule_ids] - unchangeable_delete_rule_ids)
      #opts[:delete_cond_ids] = (opts[:delete_cond_ids] - unchangeable_delete_cond_ids)
      is_unchangeable_host = (unchangeable_delete_rule_ids.any? || unchangeable_delete_cond_ids.any?)
      unchangeable_rules.select(:name).each do |rule|
        opts[:rule_messages].delete(rule.name)
      end
    end
   
    ret = {
      :delete_rule_ids => opts[:delete_rule_ids],
      :delete_cond_ids => opts[:delete_cond_ids],
      :delete_host_cond_ids => opts[:delete_host_cond_ids],
      :rule_messages => opts[:rule_messages],
      :is_unchangeable_host => is_unchangeable_host
    }
   
    ret[:hostid] = opts[:hostid] if opts[:hostid]

    ret 

  end

end
